# nmzqr.rb - libraries for Query structure and Result structure
# $Id: nmzqr.rb,v 1.2 2000/07/05 16:01:38 ryu Exp $
#
# nmzqr.rb is copyrighted free software by Ryunosuke Ohshima.
# You can use/redistribute/modify it under the terms of Namazu or Ruby.

require 'nmzqr.tab'

class Namazu
  class QueryInterface
    public
    def search()
      ResultInterface.new()
    end

    private
    def make_phrase(word)
      phrase = Array.new()
      tmpword = word.dup()
      restword = ''
      while tmpword.length() > 0
	if @nmzindex.nmzword().get_wordid(tmpword)
	  phrase.push(tmpword)
	  tmpword = restword
	  restword = ''
	else
	  tmpword.sub!(/.$/, '')
	  restword = $& + restword
	end
      end
      if restword.length() > 0
	nil
      else
	phrase
      end
    end
  end # QueryInterface

  class Query < QueryInterface
    def initialize(nmzindex, word)
      @nmzindex = nmzindex
      @word = word
    end

    public
    def search()
      phrase = make_phrase(@word)
      if phrase and phrase.length() >= 2
	results = phrase.collect() do |word|
	  Result.new(@nmzindex, word)
	end
	ResultPhrase.new(@nmzindex, results)
      else
	Result.new(@nmzindex, @word)
      end
    end
  end # Query

  class QueryPhrase < QueryInterface
    def initialize(nmzindex, words)
      @nmzindex = nmzindex
      @words = words
    end

    public
    def search()
      phrases = @words.collect() do |word|
	phrase = make_phrase(word)
	if phrase
	  phrase
	else
	  [word]
	end
      end
      results = phrases.collect() do |phrase|
	if phrase.length() >= 2
	  tmpresults = phrase.collect() do |word|
	    Result.new(@nmzindex, word)
	  end
	  ResultPhrase.new(@nmzindex, tmpresults)
	else
	  Result.new(@nmzindex, phrase[0])
	end
      end
      ResultPhrase.new(@nmzindex, results)
    end
  end # QueryPhrase

  class QueryRegexp < QueryInterface
    def initialize(nmzindex, regexpstring)
      @nmzindex = nmzindex
      @regexp = Regexp.new(regexpstring)
    end

    public
    def search()
      results = @nmzindex.nmzword().match_words(@regexp).collect() do |words|
	ResultWordid.new(@nmzindex, *words)
      end
      ResultRegexp.new(@regexp, results)
    end
  end # QueryRegexp

  class QueryMatch < QueryRegexp
    def initialize(nmzindex, matchstring)
      @nmzindex = nmzindex
      regexpstring = nil
      if /^\*(.+)\*$/ =~ matchstring
	regexpstring = $1
      elsif /^\*(.+)$/ =~ matchstring
	regexpstring = $1 + '$'
      elsif /^(.+)\*$/ =~ matchstring
	regexpstring = '^' + $1
      end
      @regexp = Regexp.new(regexpstring)
    end
  end # QueryMatch

  class QueryOperator < QueryInterface
    def initialize(queries)
      @queries = queries
    end

    public
    def search(resultope = ResultOperator)
      results = @queries.collect() do |query|
	query.search()
      end
      resultope.new(results)
    end
  end # QueryOperator

  class QueryAnd < QueryOperator
    public
    def search()
      super(ResultAnd)
    end
  end # QueryAnd

  class QueryOr < QueryOperator
    public
    def search()
      super(ResultOr)
    end
  end # QueryOr

  class QueryNot < QueryOperator
    public
    def search()
      super(ResultNot)
    end
  end # QueryNot

  class ResultInterface
    def initialize()
      @scoreresult = ScoreResult.new()
    end
    attr_reader(:scoreresult)

    public
    def word()
      ""
    end

    public
    def docids()
      @scoreresult.docids()
    end

    public
    def hitnum()
      @scoreresult.length()
    end

    public
    def score(docid)
      @scoreresult.score(docid)
    end

    public
    def time(docid)
      @scoreresult.time(docid)
    end

    public
    def to_s()
      '[' + word() + ': ' + hitnum().to_s() + ']'
    end
  end # ResultInterface

  class Result < ResultInterface
    def initialize(nmzindex, word)
      @word = word
      @scoreresult = nmzindex.nmzscore().get_scores_by_wordid(nmzindex.nmzword().get_wordid(word))
      @idf = Math::log(nmzindex.docnum() / hitnum()) / Math::log(2)
    end
    attr_reader(:word)

    public
    def score(docid)
      score = @scoreresult.score(docid)
      if score
	score * @idf + 1
      else
	0
      end
    end
  end # Result

  class ResultWordid < Result
    def initialize(nmzindex, word, wordid)
      @word = word
      @scoreresult = nmzindex.nmzscore().get_scores_by_wordid(wordid)
      @idf = Math::log(nmzindex.docnum() / hitnum()) / Math::log(2)
    end
    attr_reader(:word)
  end # ResultWordid

  class ResultOperator < ResultInterface
    attr_reader(:results)

    def word()
      @results.collect() do |result| result.word() end .join('')
    end

    def score(docid)
      score = 0
      for result in @results
	score += result.score(docid)
      end
      score
    end

    def time(docid)
      time = nil
      @results.each() do |result|
	time = result.time(docid)
	break if time
      end
      time
    end

    def to_s()
      '[' + @results.collect() { |result| result.to_s() } .join(' ') + ' :: ' + hitnum().to_s() + ']'
    end
  end # ResultOperator

  class ResultPhrase < ResultOperator
    def initialize(nmzindex, results)
      @results = results
      @scoreresult = nil
      if @results.length() > 0
	@scoreresult = @results[0].scoreresult().clone()
	for i in 1 .. @results.length() - 1
	  tmpscoreresult = @results[i].scoreresult()
	  @scoreresult.each() do |docid, oldvalue|
	    newvalue = tmpscoreresult[docid]
	    if (!newvalue)
	      @scoreresult.delete(docid)
	    elsif (oldvalue[0] < newvalue[0])
	      @scoreresult[docid] = newvalue
	    end
	  end
	end

	phrasestring = @results.collect() do |result| result.word() end .join('')
	phrasedocids = nmzindex.nmzphrase().get_docids(phrasestring)
	@scoreresult.delete_if() do |docid, value|
	  !phrasedocids.include?(docid)
	end
      else
	@scoreresult = ScoreResult.new()
      end
    end

    def score(docid)
      @scoreresult.score(docid)
    end

    def to_s()
      '{' + @results.collect() { |result| result.to_s() } .join(' ') + ' :: ' + hitnum().to_s() + '}'
    end
  end # ResultPhrase

  class ResultRegexp < ResultOperator
    def initialize(regexp, results)
      @word = regexp.inspect()
      @results = results
      @scoreresult = nil
      if @results.length() > 0
	@scoreresult = @results[0].scoreresult().clone()
	for i in 1 .. @results.length() - 1
	  tmpscoreresult = @results[i].scoreresult()
	  tmpscoreresult.each() do |docid, newvalue|
	    oldvalue = @scoreresult[docid]
	    if (!oldvalue or oldvalue[0] < newvalue[0])
	      @scoreresult[docid] = newvalue
	    end
	  end
	end
      else
	@scoreresult =	ScoreResult.new()
      end
    end
    attr_reader(:word)

    def score(docid)
      @scoreresult.score(docid)
    end

    public
    def to_s()
      '{' + word() + ' :: ' + hitnum().to_s() + '}'
    end
  end # ResultRegexp

  class ResultAnd < ResultOperator
    def initialize(results)
      @results = results
      @scoreresult = nil
      if @results.length() > 0
	@scoreresult = @results[0].scoreresult().clone()
	for i in 1 .. @results.length() - 1
	  tmpscoreresult = @results[i].scoreresult()
	  @scoreresult.delete_if() do |docid, value|
	    !tmpscoreresult[docid]
	  end
	end
      else
	@scoreresult =	ScoreResult.new()
      end
    end

    def to_s()
      '{' + @results.collect() { |result| result.to_s() } .join(' & ') + ' :: ' + hitnum().to_s() + '}'
    end
  end # ResultAnd

  class ResultOr < ResultOperator
    def initialize(results)
      @results = results
      @scoreresult = nil
      if @results.length() > 0
	@scoreresult = @results[0].scoreresult().clone()
	for i in 1 .. @results.length() - 1
	  tmpscoreresult = @results[i].scoreresult()
	  tmpscoreresult.each() do |docid, value|
	    if @scoreresult[docid]
	      next
	    end
	    @scoreresult[docid] = value
	  end
	end
      else
	@scoreresult =	ScoreResult.new()
      end
    end

    def to_s()
      '{' + @results.collect() { |result| result.to_s() } .join(' | ') + ' :: ' + hitnum().to_s() + '}'
    end
  end # ResultOr

  class ResultNot < ResultOperator
    def initialize(results)
      @results = results
      if @results.length() > 0
	@scoreresult = @results[0].scoreresult().clone()
	for i in 1 .. @results.length() - 1
	  tmpscoreresult = @results[i].scoreresult()
	  @scoreresult.delete_if() do |docid, value|
	    tmpscoreresult[docid]
	  end
	end
      else
	@scoreresult =	ScoreResult.new()
      end
    end

    def score(docid)
      @results[0] ? @results[0].score(docid) : 0
    end

    def to_s()
      '{' + @results.collect() { |result| result.to_s() } .join(' - ') + ' :: ' + hitnum().to_s() + '}'
    end
  end # ResultNot
end # Namazu
